package org.roboquant.brokers.sim

import org.roboquant.feeds.Event
import org.roboquant.orders.*
import java.util.*
import kotlin.reflect.KClass

/**
 * Order handler factory creates [OrderHandler] for an order. The provided handler is repsonsible for executing the
 * order.
 *
 * @param T
 * @constructor Create empty Order handler factory
 */
fun interface OrderHandlerFactory<T : Order> {

    /**
     * Get a handler for the provided order
     *
     * @param order
     * @return
     */
    fun getHandler(order: T): OrderHandler
}

/**
 * Engine that simulates how orders are executed on financial markets. For any order to be executed, it needs a
 * corresponding [OrderHandler]
 *
 * @property pricingEngine
 * @constructor Create empty Execution engine
 */
class ExecutionEngine(private val pricingEngine: PricingEngine = NoCostPricingEngine()) {

    companion object {

        /**
         * All the registered [OrderHandlerFactory]
         */
        val factories = mutableMapOf<KClass<*>, OrderHandlerFactory<Order>>()

        /**
         * Get the order handler for the provided [order]
         *
         * @param order
         * @return
         */
        fun getHandler(order: Order): OrderHandler {
            val factory = factories[order::class]!!
            return factory.getHandler(order)
        }

        /**
         * Unregister the order handler for order type [T]
         */
        inline fun <reified T : Order> unregister() {
            factories.remove(T::class)
        }

        /**
         * Register a new order handler factory for order type [T]. If there was already an order handler registered
         * for the same class it will be overwritten.
         *
         * @param T
         * @param factory
         */
        inline fun <reified T : Order> register(factory: OrderHandlerFactory<T>) {
            @Suppress("UNCHECKED_CAST")
            factories[T::class] = factory as OrderHandlerFactory<Order>
        }

        init {
            // register all the default included order handlers

            // Single Order types
            register<MarketOrder> { MarketOrderHandler(it) }
            register<LimitOrder> { LimitOrderHandler(it) }
            register<StopLimitOrder> { StopLimitOrderHandler(it) }
            register<StopOrder> { StopOrderHandler(it) }
            register<TrailLimitOrder> { TrailLimitOrderHandler(it) }
            register<TrailOrder> { TrailOrderHandler(it) }

            // Advanced order types
            register<BracketOrder> { BracketOrderHandler(it) }
            register<OCOOrder> { OCOOrderHandler(it) }
            register<OTOOrder> { OTOOrderHandler(it) }

            // Modify order types
            register<UpdateOrder> { UpdateOrderHandler(it) }
            register<CancelOrder> { CancelOrderHandler(it) }
        }

    }


    // Currently active order commands
    private val tradeHandlers = LinkedList<TradeOrderHandler>()

    // Currently active order commands
    private val modifyHandlers = LinkedList<ModifyOrderHandler>()

    /**
     * Get the open order handlers
     */
    private fun <T: OrderHandler>List<T>.open() = filter { it.status.open }

    /**
     * Remove all handlers of closed orders
     */
    internal fun removeClosedOrders() {
        tradeHandlers.removeIf { it.state.status.closed }
        modifyHandlers.removeIf { it.state.status.closed }
    }


    /**
     * Latetst Order states of all handlers
     */
    val orderStates
        get() = tradeHandlers.map { it.state } + modifyHandlers.map { it.state }

    /**
     * Add a new order to the execution engine. Orders can only be processed if their is a corresponding handler
     * registered for the order class.
     *
     * @param order
     * @return
     */
    fun add(order: Order): Boolean {

        return when (val handler = getHandler(order)) {
            is ModifyOrderHandler -> modifyHandlers.add(handler)
            is TradeOrderHandler -> tradeHandlers.add(handler)
        }

    }


    // Add a new order to the execution engine
    fun addAll(orders: List<Order>) {
        for (order in orders) add(order)
    }

    /**
     * Execute all the handlers of orders that are not yet closed. Logic:
     *
     * 1. First any open modify orders will be processed
     * 2. The any regular order will be processed assuming there is a price action for the underlying asses.
     *
     * @param event
     * @return
     */
    fun execute(event: Event): List<Execution> {

        // We always first execute the modify orders. These are run even if there is
        // no price for the asset known
        for (handler in modifyHandlers.open()) {
            handler.execute(tradeHandlers, event.time)
        }

        // Now run the trade order commands
        val executions = mutableListOf<Execution>()
        val prices = event.prices
        for (handler in tradeHandlers.open()) {
            val action = prices[handler.state.asset] ?: continue
            val pricing = pricingEngine.getPricing(action, event.time)
            val newExecutions = handler.execute(pricing, event.time)
            executions.addAll(newExecutions)
        }
        return executions
    }

    /**
     * Clear any state
     */
    fun clear() {
        tradeHandlers.clear()
        pricingEngine.clear()
    }

}


